// FileStreams_functions.js - 2011-11-19
// -    NTFS 
// based on the article: http://www.flexhex.com/docs/articles/alternate-streams.phtml
//
// Contains functions:
// IsSupportFileStreams()
// IsFileStreamExists()
// CreateFileStream()
// DeleteFileStream()
// WriteFileStream()
// EnumFileStreams()
//
// Usage:
// AkelPad.Include("FileStreams_functions.js");

//------------------------------------------------------------
// bSupport = IsSupportFileStreams(sDrive)
//
// Argument:
// sDrive - drive where the file is saved, eg. "D:\\"
//
// Return value:
// true if file system supports NTFS streams, otherwise false.
//------------------------------------------------------------
function IsSupportFileStreams(sDrive)
{
  var lpBuffer = AkelPad.MemAlloc(4);
  var bSupport = false;

  AkelPad.SystemFunction().Call("kernel32::GetVolumeInformation" + _TCHAR, sDrive, 0, 0, 0, 0, lpBuffer, 0, 0);

  if (AkelPad.MemRead(lpBuffer, 3 /*DT_DWORD*/) & 0x00040000 /*FILE_NAMED_STREAMS*/)
    bSupport = true;

  AkelPad.MemFree(lpBuffer);

  return bSupport;
}

//--------------------------------------------------
// bIsStream = IsFileStreamExists(sFile, sStream)
//
// Arguments:
// sFile   - full file name, eg. "D:\\Text\\abc.txt"
// sStream - stream name
//
// Return value:
// true if sFile contains sStream, otherwise false.
//--------------------------------------------------
function IsFileStreamExists(sFile, sStream)
{
  var oSys    = AkelPad.SystemFunction();
  var hStream = oSys.Call("kernel32::CreateFile" + _TCHAR,
                          sFile + ":" + sStream, //lpFileName
                          0,  //dwDesiredAccess
                          3,  //dwShareMode = FILE_SHARE_READ|FILE_SHARE_WRITE
                          0,  //lpSecurityAttributes
                          3,  //dwCreationDisposition = OPEN_EXISTING
                          0,  //dwFlagsAndAttributes
                          0); //hTemplateFile

  if (hStream != -1) //INVALID_HANDLE_VALUE
  {
    oSys.Call("kernel32::CloseHandle", hStream);
    return true;
  }
  else
    return false;
}

//--------------------------------------------------
// bCreated = CreateFileStream(sFile, sStream)
//
// Arguments:
// sFile   - full file name, eg. "D:\\Text\\abc.txt"
// sStream - stream name
//
// Return value:
// true if sStream was created, otherwise false.
//--------------------------------------------------
function CreateFileStream(sFile, sStream)
{
  var oSys    = AkelPad.SystemFunction();
  var hStream = oSys.Call("kernel32::CreateFile" + _TCHAR,
                          sFile + ":" + sStream, //lpFileName
                          0,  //dwDesiredAccess
                          3,  //dwShareMode = FILE_SHARE_READ|FILE_SHARE_WRITE
                          0,  //lpSecurityAttributes
                          1,  //dwCreationDisposition = CREATE_NEW
                          0,  //dwFlagsAndAttributes
                          0); //hTemplateFile

  if (hStream != -1) //INVALID_HANDLE_VALUE
  {
    oSys.Call("kernel32::CloseHandle", hStream);
    return true;
  }
  else
    return false;
}

//--------------------------------------------------
// bDeleted = DeleteFileStream(sFile, sStream)
//
// Arguments:
// sFile   - full file name, eg. "D:\\Text\\abc.txt"
// sStream - stream name
//
// Return value:
// true if sStream was deleted, otherwise false.
//--------------------------------------------------
function DeleteFileStream(sFile, sStream)
{
  if (AkelPad.SystemFunction().Call("kernel32::DeleteFile" + _TCHAR, sFile + ":" + sStream))
    return true;
  else
    return false;
}

//---------------------------------------------------------
// bWritten = WriteFileStream(sFile, sStream, sData, nType)
//
// Arguments:
// sFile   - full file name, eg. "D:\\Text\\abc.txt"
// sStream - stream name
// sData   - text data to write
// nType   - type of the sData parameter:
//           0 - ANSI
//           1 - UNICODE
//
// Return value:
// bWritten - true if the function succeeds.
//---------------------------------------------------------
function WriteFileStream(sFile, sStream, sData, nType)
{
  var oSys     = AkelPad.SystemFunction();
  var bWritten = false;
  var hStream  = oSys.Call("kernel32::CreateFile" + _TCHAR,
                           sFile + ":" + sStream, //lpFileName
                           0x40000000, //dwDesiredAccess = GENERIC_WRITE
                           3,          //dwShareMode = FILE_SHARE_READ|FILE_SHARE_WRITE
                           0,          //lpSecurityAttributes
                           5,          //dwCreationDisposition = TRUNCATE_EXISTING
                           0,          //dwFlagsAndAttributes
                           0);         //hTemplateFile
  var nBufLen;
  var lpBuffer;
  var lpNumBytes;

  if (hStream != -1) //INVALID_HANDLE_VALUE
  {
    nBufLen = sData.length + 1;
    if (nType)
      nBufLen = 2 + nBufLen * 2;

    lpBuffer   = AkelPad.MemAlloc(nBufLen);
    lpNumBytes = AkelPad.MemAlloc(4);

    if (nType)
    {
      //BOM UTF-16LE
      AkelPad.MemCopy(lpBuffer,     0xFF, 5 /*DT_BYTE*/);
      AkelPad.MemCopy(lpBuffer + 1, 0xFE, 5 /*DT_BYTE*/);
      AkelPad.MemCopy(lpBuffer + 2, sData, nType);
    }
    else
      AkelPad.MemCopy(lpBuffer, sData, nType);

    bWritten = oSys.Call("kernel32::WriteFile",
                         hStream,             //hFile
                         lpBuffer,            //lpBuffer
                         nBufLen - 1 - nType, //nNumberOfBytesToWrite
                         lpNumBytes,          //lpNumberOfBytesWritten
                         0);                  //lpOverlapped
 
    oSys.Call("kernel32::CloseHandle", hStream);
    AkelPad.MemFree(lpBuffer);
    AkelPad.MemFree(lpNumBytes);
  }

  return bWritten;
}

//--------------------------------------------------------------
// aStream = EnumFileStreams(sFile)
//
// Argument:
// sFile - full file name, eg. "D:\\Text\\abc.txt"
//
// Return value:
// Array. Each element of array is two-elements array, contains information about a single file stream:
// aStream[n][0] - file stream name,
// aStream[n][1] - file stream size in bytes.
//
//The function can only be applied to files, not to directories.
//--------------------------------------------------------------
function EnumFileStreams(sFile)
{
  var oSys    = AkelPad.SystemFunction();
  var aStream = [];
  var nVerLen = 4 * 5 + 128 * _TSIZE;
  var lpOsVer = AkelPad.MemAlloc(nVerLen);
  var nMajorVer;
  var nMinorVer;
  var sStreamName;
  var nStreamSize;

  //Get system version
  AkelPad.MemCopy(lpOsVer, nVerLen, 3 /*DT_DWORD*/);
  oSys.Call("kernel32::GetVersionEx" + _TCHAR, lpOsVer);
  nMajorVer = AkelPad.MemRead(lpOsVer + 4, 3 /*DT_DWORD*/);
  nMinorVer = AkelPad.MemRead(lpOsVer + 8, 3 /*DT_DWORD*/);
  AkelPad.MemFree(lpOsVer);

  //Win-Vista +, Win-Server-2003 +
  if ((nMajorVer > 5) || ((nMajorVer == 5) && (nMinorVer >= 2)))
  {
    var lpFindStream = AkelPad.MemAlloc(8 + (260 + 36) * 2);
    var hFindStream  = oSys.Call("kernel32::FindFirstStreamW", sFile, 0 /*FindStreamInfoStandard*/, lpFindStream, 0);

    if (hFindStream != -1) //INVALID_HANDLE_VALUE
    {
      do
      {
        sStreamName = AkelPad.MemRead(lpFindStream + 8, _TSTR);
        nStreamSize = GetFileStreamSize(lpFindStream);

        aStream[aStream.length] = [sStreamName, nStreamSize];
      }
      while(oSys.Call("kernel32::FindNextStreamW", hFindStream, lpFindStream));
    }

    oSys.Call("kernel32::FindClose", hFindStream);
    AkelPad.MemFree(lpFindStream);
  }

  //Win-XP
  else
  {
    var nInfoLen = 65536;
    var lpInfo   = AkelPad.MemAlloc(nInfoLen);
    var lpStatus = AkelPad.MemAlloc(8);
    var hFile;
    var lpElement;
    var nNextOffset;
    var nNameLength;
    var i;

    hFile = oSys.Call("kernel32::CreateFile" + _TCHAR,
                      sFile, //lpFileName
                      0,     //dwDesiredAccess
                      3,     //dwShareMode = FILE_SHARE_READ|FILE_SHARE_WRITE
                      0,     //lpSecurityAttributes
                      3,     //dwCreationDisposition = OPEN_EXISTING
                      0,     //dwFlagsAndAttributes
                      0);    //hTemplateFile

    if (hFile != -1) //INVALID_HANDLE_VALUE
    {
      oSys.Call("ntdll::NtQueryInformationFile",
                hFile,    //FileHandle
                lpStatus, //IoStatusBlock
                lpInfo,   //FileInformation
                nInfoLen, //Length
                22);      //FileInformationClass:FileStreamInformation
                // http://msdn.microsoft.com/en-us/library/cc232090%28v=prot.13%29.aspx

      lpElement   = lpInfo;
      nNextOffset = 0;

      do
      {
        lpElement  += nNextOffset;
        nNameLength = AkelPad.MemRead(lpElement + 4, 3 /*DT_DWORD*/);

        sStreamName = "";
        for (i = 24; i < 24 + nNameLength; i = i + 2)
          sStreamName += String.fromCharCode(AkelPad.MemRead(lpElement + i, 4 /*DT_WORD*/));

        sStreamName = sStreamName.substring(1, sStreamName.lastIndexOf(":"));
        nStreamSize = GetFileStreamSize(lpElement + 8);

        aStream[aStream.length] = [sStreamName, nStreamSize];
      }
      while (nNextOffset = AkelPad.MemRead(lpElement, 3 /*DT_DWORD*/));

      oSys.Call("kernel32::CloseHandle", hFile);
    }

    AkelPad.MemFree(lpInfo);
    AkelPad.MemFree(lpStatus);
  }

  return aStream;
}

function GetFileStreamSize(lpBuffer)
{
  var sNum = "";
  var sByte;
  var i;

  for (i = 0; i < 8; ++i)
  {
    sByte = AkelPad.MemRead(lpBuffer + i, 5 /*DT_BYTE*/).toString(16);
    if (sByte.length == 1)
      sByte = "0" + sByte;

    sNum = sByte + sNum;
  }

  return parseInt("0x" + sNum);
}
