/// quoted-printable.js
// http://akelpad.sourceforge.net/forum/viewtopic.php?p=32125#32125
// Version: 1.0.1 (2016-12-02)
// Author: Mathias Bynens - original version, Andrey Tzar - adaptation for AkelPad editor
// Origin: https://github.com/mathiasbynens/quoted-printable
// License: MIT https://mths.be/mit

// Description(1033): Decode/encode quoted-printable strings (selected in AkelPad editor)
// Description(1049): / quoted-printable    

/*
   Argument:
      First argument:   'decode' or 'encode' - do decode or encode operation on selected string
      Default is: `decode`

   Example:
      Call("Scripts::Main", 1, "quoted-printable.js", `encode`) - encode current selection into quoted-printable string
   
   Note: operation performs on the spot

   :
        - 'decode' -      quoted-printable    (  , ..       ),
         'encode' -      quoted-printable

   :
      Call("Scripts::Main", 1, "quoted-printable.js", `encode`);
         foo=0D=0Abar => foo=3D0D=3D0Abar
         ! => =D0=9F=D1=80=D0=B8=D0=B2=D0=B5=D1=82!
*/

/*! https://mths.be/quoted-printable v1.0.1 by @mathias | MIT license */
var quotedPrintable = (function() {
   function convertToUnicode(str, cp) {
      var oSys = AkelPad.SystemFunction();

      if(cp === undefined)
         cp = AkelPad.GetEditCodePage(0);
   
      var ret = "";
   
      if(
         cp == 1200 //UTF-16LE
         || cp == 1201 //UTF-16BE
         || cp == 12000 //UTF-32LE
         || cp == 12001 //UTF-32BE
      ) {
         var isLE = cp == 1200 || cp == 12000;
         var step = cp == 12000 || cp == 12001 ? 4 : 2;
         if(str.length % step != 0)
            throw "Invalid unicode string";
         for(var i = 0, l = str.length; i < l; i += step) {
            var chars = str.substr(i, step);
            if(isLE) {
               var b1 = chars.charCodeAt(0);
               var b2 = chars.charCodeAt(1);
            }
            else {
               var b1 = chars.charCodeAt(step - 1);
               var b2 = chars.charCodeAt(step - 2);
            }
            ret += String.fromCharCode((b2 << 8) + b1);
         }
         return ret;
      }
   
      // based on Fr0sT's code: http://akelpad.sourceforge.net/forum/viewtopic.php?p=7972#7972
   
      // current code page is UTF16* or UTF32* - set ansi current code page
      // (WideChar <-> MultiByte functions don't work with this code pages)
      if(cp == 1 || cp == 1200 || cp == 1201 || cp == 12000 || cp == 12001)
         cp = 0;
   
      try {
         var strLen = str.length;
         var pMBBuf = AkelPad.MemAlloc(strLen * 1 /*sizeof(char)*/);
         if(!pMBBuf)
            throw new Error("MemAlloc fail");
         for(var i = 0; i < strLen; ++i)
            AkelPad.MemCopy(_PtrAdd(pMBBuf, i), str.charCodeAt(i), 5 /*DT_BYTE*/);
   
         // get required buffer size
         var bufLen = oSys.Call(
            "Kernel32::MultiByteToWideChar",
            cp,       //   __in   UINT CodePage,
            0,        //   __in   DWORD dwFlags,
            pMBBuf,   //   __in   LPCSTR lpMultiByteStr,
            strLen,   //   __in   int cbMultiByte,
            0,        //   __out  LPWSTR lpWideCharStr,
            0         //   __in   int cchWideChar
         );
         if(!bufLen)
            throw new Error("MultiByteToWideChar fail " + oSys.GetLastError());
         bufLen *= 2 /*sizeof(wchar_t)*/;
   
         var pWCBuf = AkelPad.MemAlloc(bufLen);
         if(!pWCBuf)
            throw new Error("MemAlloc fail");
   
         // convert buffer
         bufLen = oSys.Call(
            "Kernel32::MultiByteToWideChar",
            cp,       //   __in   UINT CodePage,
            0,        //   __in   DWORD dwFlags,
            pMBBuf,   //   __in   LPCSTR lpMultiByteStr,
            strLen,   //   __in   int cbMultiByte,
            pWCBuf,   //   __out  LPWSTR lpWideCharStr,
            bufLen    //   __in   int cchWideChar
         );
         if(!bufLen)
            throw new Error("MultiByteToWideChar fail " + oSys.GetLastError());
   
         //ret = AkelPad.MemRead(pWCBuf, 1 /*DT_UNICODE*/);
         for(
            var pCurr = pWCBuf, bufMax = _PtrAdd(pWCBuf, bufLen*2);
            _PtrMath(pCurr, "<", bufMax);
            pCurr = _PtrAdd(pCurr, 2)
         )
            ret += String.fromCharCode(AkelPad.MemRead(pCurr, 4 /*DT_WORD*/));
      }
      catch(e) {
         throw e;
      }
      finally {
         pMBBuf && AkelPad.MemFree(pMBBuf);
         pWCBuf && AkelPad.MemFree(pWCBuf);
      }
   
      return ret;
   }

   function convertFromUnicode(str, cp) {
      var oSys = AkelPad.SystemFunction();

      if(cp === undefined)
         cp = AkelPad.GetEditCodePage(0);
   
      var ret = "";
   
      if(
         cp == 1200 //UTF-16LE
         || cp == 1201 //UTF-16BE
         || cp == 12000 //UTF-32LE
         || cp == 12001 //UTF-32BE
      ) {
         var isLE = cp == 1200 || cp == 12000;
         var u32 = cp == 12000 || cp == 12001 ? "\x00\x00" : "";
         for(var i = 0, l = str.length; i < l; ++i) {
            var code = str.charCodeAt(i);
            var b1 = String.fromCharCode(code & 0xff);
            var b2 = String.fromCharCode(code >> 8 & 0xff);
            ret += isLE
               ? b1 + b2 + u32
               : u32 + b2 + b1;
         }
         return ret;
      }
   
      // based on Fr0sT's code: http://akelpad.sourceforge.net/forum/viewtopic.php?p=7972#7972
   
      // current code page is UTF16* or UTF32* - set ansi current code page
      // (WideChar <-> MultiByte functions don't work with this code pages)
      if(cp == 1 || cp == 1200 || cp == 1201 || cp == 12000 || cp == 12001)
         cp = 0;
   
      try {
         var strLen = str.length + 1;
         var pWCBuf = AkelPad.MemAlloc(strLen * _TSIZE);
         if(!pWCBuf)
            throw new Error("MemAlloc fail");
         AkelPad.MemCopy(pWCBuf, str, _TSTR);
         
         // get required buffer size
         var bufLen = oSys.Call(
            "Kernel32::WideCharToMultiByte",
            cp,       //   __in   UINT CodePage,
            0,        //   __in   DWORD dwFlags,
            pWCBuf,   //   __in   LPCWSTR lpWideCharStr,
            strLen,   //   __in   int cchWideChar,
            0,        //   __out  LPSTR lpMultiByteStr,
            0,        //   __in   int cbMultiByte,
            0,        //   __in   LPCSTR lpDefaultChar,
            0         //   __out  LPBOOL lpUsedDefaultChar
         );
         if(!bufLen)
            throw new Error("WideCharToMultiByte fail " + oSys.GetLastError());
         
         var pMBBuf = AkelPad.MemAlloc(bufLen);
         if(!pMBBuf)
            throw new Error("MemAlloc fail");
   
         // convert buffer
         bufLen = oSys.Call(
            "Kernel32::WideCharToMultiByte",
            cp,       //   __in   UINT CodePage,
            0,        //   __in   DWORD dwFlags,
            pWCBuf,   //   __in   LPCWSTR lpWideCharStr,
            strLen,   //   __in   int cchWideChar,
            pMBBuf,   //   __out  LPSTR lpMultiByteStr,
            bufLen,   //   __in   int cbMultiByte,
            0,        //   __in   LPCSTR lpDefaultChar,
            0         //   __out  LPBOOL lpUsedDefaultChar
         );
         if(!bufLen)
            throw new Error("WideCharToMultiByte fail " + oSys.GetLastError());
   
         //ret = AkelPad.MemRead(pMBBuf, 0 /*DT_ANSI*/);
         for(
            var pCurr = pMBBuf, bufMax = _PtrAdd(pMBBuf, bufLen - 1);
            _PtrMath(pCurr, "<", bufMax);
            pCurr = _PtrAdd(pCurr, 1)
         )
            ret += String.fromCharCode(AkelPad.MemRead(pCurr, 5 /*DT_BYTE*/));
      }
      catch(e) {
         throw e;
      }
      finally {
         pWCBuf && AkelPad.MemFree(pWCBuf);
         pMBBuf && AkelPad.MemFree(pMBBuf);
      }
   
      return ret;
   }

   /*--------------------------------------------------------------------------*/

   var stringFromCharCode = String.fromCharCode;

   var decode = function(input) {
      var s = input
         // https://tools.ietf.org/html/rfc2045#section-6.7, rule 3:
         // Therefore, when decoding a `Quoted-Printable` body, any trailing white
         // space on a line must be deleted, as it will necessarily have been added
         // by intermediate transport agents.
         .replace(/[\t\x20]$/gm, '')
         // Remove hard line breaks preceded by `=`. Proper `Quoted-Printable`-
         // encoded data only contains CRLF line  endings, but for compatibility
         // reasons we support separate CR and LF too.
         .replace(/=(?:\r\n?|\n|$)/g, '')
         // Decode escape sequences of the form `=XX` where `XX` is any
         // combination of two hexidecimal digits. For optimal compatibility,
         // lowercase hexadecimal digits are supported as well. See
         // https://tools.ietf.org/html/rfc2045#section-6.7, note 1.
         .replace(/=([a-fA-F0-9]{2})/g, function($0, $1) {
            var codePoint = parseInt($1, 16);
            return stringFromCharCode(codePoint);
         });
      // WScript.Echo(s);
      return convertToUnicode(s, 65001);
      
   };

   var handleTrailingCharacters = function(string) {
      return string
         .replace(/\x20$/, '=20') // Handle trailing space.
         .replace(/\t$/, '=09') // Handle trailing tab.
   };

   var regexUnsafeSymbols = /[\0-\x08\n-\x1F=\x7F-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]/g;
   
   var encode = function(string) {
      string = convertFromUnicode(string, 65001);
      // Encode symbols that are definitely unsafe (i.e. unsafe in any context).
      var encoded = string.replace(regexUnsafeSymbols, function(symbol) {
         if (symbol > '\xFF') {
            throw RangeError(
               '`quotedPrintable.encode()` expects extended ASCII input only. ' +
               'Don\u2019t forget to encode the input first using a character ' +
               'encoding like UTF-8.'
            );
         }
         var codePoint = symbol.charCodeAt(0);
         var hexadecimal = codePoint.toString(16).toUpperCase();
         return '=' + ('0' + hexadecimal).slice(-2);
      });

      // Limit lines to 76 characters (not counting the CRLF line endings).
      var lines = encoded.split(/\r\n?|\n/g);
      var lineIndex = -1;
      var lineCount = lines.length;
      var result = [];
      while (++lineIndex < lineCount) {
         var line = lines[lineIndex];
         // Leave room for the trailing `=` for soft line breaks.
         var LINE_LENGTH = 75;
         var index = 0;
         var length = line.length;
         while (index < length) {
            var buffer = encoded.slice(index, index + LINE_LENGTH);
            // If this line ends with `=`, optionally followed by a single uppercase
            // hexadecimal digit, we broke an escape sequence in half. Fix it by
            // moving these characters to the next line.
            if (/=$/.test(buffer)) {
               buffer = buffer.slice(0, LINE_LENGTH - 1);
               index += LINE_LENGTH - 1;
            } else if (/=[A-F0-9]$/.test(buffer)) {
               buffer = buffer.slice(0, LINE_LENGTH - 2);
               index += LINE_LENGTH - 2;
            } else {
               index += LINE_LENGTH;
            }
            result.push(buffer);
         }
      }

      // Encode space and tab characters at the end of encoded lines. Note that
      // with the current implementation, this can only occur at the very end of
      // the encoded string  every other line ends with `=` anyway.
      var lastLineLength = buffer.length;
      if (/[\t\x20]$/.test(buffer)) {
         // There's a space or a tab at the end of the last encoded line. Remove
         // this line from the `result` array, as it needs to change.
         result.pop();
         if (lastLineLength + 2 <= LINE_LENGTH + 1) {
            // It's possible to encode the character without exceeding the line
            // length limit.
            result.push(
               handleTrailingCharacters(buffer)
            );
         } else {
            // It's not possible to encode the character without exceeding the line
            // length limit. Remvoe the character from the line, and insert a new
            // line that contains only the encoded character.
            result.push(
               buffer.slice(0, lastLineLength - 1),
               handleTrailingCharacters(
                  buffer.slice(lastLineLength - 1, lastLineLength)
               )
            );
         }
      }

      // `Quoted-Printable` uses CRLF.
      var s = result.join('=\r\n');
      return s;
   };

   return {
      'encode': encode,
      'decode': decode,
      'version': '1.0.1'
   };


}());

/* AkelPad specific code below */
// check if we use this script out of AkelPad editor's environment
if (typeof AkelPad == "undefined")
{
  try
  {
    AkelPad = new ActiveXObject("AkelPad.Document");
    _PtrAdd = function(n1, n2) {return AkelPad.Global._PtrAdd(n1, n2);};
    _X64    = AkelPad.Constants._X64;
    AkelPad.ActiveXThis(this);
  }
  catch (oError)
  {
    WScript.Echo("You need to register Scripts.dll");
    WScript.Quit();
  }
}

// find selected text and throw error if there is no one
var nSelStart = AkelPad.GetSelStart(),
      nSelEnd = AkelPad.GetSelEnd();
var inputBuf = trim(AkelPad.GetSelText());
if(nSelStart == nSelEnd || inputBuf.length < 1)
   error("Select text to operate");

var option = (WScript.Arguments.length ? trim(WScript.Arguments(0)) : 'decode');
var text = option == 'encode' ? quotedPrintable.encode(inputBuf) :
   quotedPrintable.decode(inputBuf);
// var trailing = option == 'encode' ? 1 : 0;

AkelPad.ReplaceSel(text);
AkelPad.SetSel(nSelStart, nSelStart + text.length /* - trailing */ );

/* ******************* */
/* Auxiliary functions */
/* ******************* */

// trim spaces from beginning and end of a string
function trim(pStr)
{
  return (typeof pStr == 'string' && pStr.length) ?
     pStr.replace(/^\s+|\s+$/, '') : '';
}

// error handling
function error(s)
{
      AkelPad.MessageBox(AkelPad.GetEditWnd(), s, WScript.ScriptName, 0x48);
      WScript.Quit();   
}
